from airthings_ble.const import CO2_MAX, PERCENTAGE_MAX, PRESSURE_MAX, RADON_MAX
from airthings_ble.parser import illuminance_converter, validate_value


def test_validate_value_humidity() -> None:
    valid_humidity_values = [0, 50, 100.0]
    for value in valid_humidity_values:
        assert validate_value(value=value, max_value=PERCENTAGE_MAX) == value

    invalid_humidity_values = [-1, 100.1, 101]
    for value in invalid_humidity_values:
        assert validate_value(value=value, max_value=PERCENTAGE_MAX) is None


def test_validate_value_radon() -> None:
    valid_radon_values = [0, 100, 1000.0, 16383]
    for value in valid_radon_values:
        assert validate_value(value=value, max_value=RADON_MAX) == value

    invalid_radon_values = [-1, 16384, 65535]
    for value in invalid_radon_values:
        assert validate_value(value=value, max_value=RADON_MAX) is None


def test_validate_value_co2() -> None:
    valid_co2_values = [0, 100, 1000.0, 65534]
    for value in valid_co2_values:
        assert validate_value(value=value, max_value=CO2_MAX) == value

    invalid_co2_values = [-1, 65535]
    for value in invalid_co2_values:
        assert validate_value(value=value, max_value=CO2_MAX) is None


def test_validate_value_illuminance() -> None:
    assert illuminance_converter(0) == 0
    assert illuminance_converter(255) == 100
    assert illuminance_converter(256) is None


def test_validata_value_pressure() -> None:
    assert validate_value(value=0.0, max_value=PRESSURE_MAX) == 0
    assert validate_value(value=1310.0, max_value=PRESSURE_MAX) == 1310
    assert validate_value(value=1311.0, max_value=PRESSURE_MAX) is None
    assert validate_value(value=-1.0, max_value=PRESSURE_MAX) is None
    assert validate_value(value=65535.0, max_value=PRESSURE_MAX) is None



from airthings_ble.parser import _decode_wave_illum_accel


def test_wave_gen_1_illuminance_and_accelerometer() -> None:
    """Test Wave Gen 1 illuminance and accelerometer."""
    raw_data = bytearray.fromhex("b20c")

    decoded_data = _decode_wave_illum_accel(
        name="illuminance_accelerometer", format_type="BB", scale=1.0
    )(raw_data)

    assert decoded_data["illuminance"] == 69



import logging

from airthings_ble.atom.request_path import AtomRequestPath
from airthings_ble.atom.response import AtomResponse

_LOGGER = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)


def test_atom_response_wave_enhance() -> None:
    """Test the Wave Enhance request."""
    random_bytes = bytes.fromhex("A1B2")

    response = AtomResponse(
        logger=_LOGGER,
        response=bytes.fromhex(
            "1001000345a1b281a2006d32393939392f302f333130313202583ea9634e4f49"
            + "182763544d501972f06348554d190d2f63434f321902dc63564f43190115634c5"
            + "55801635052531a005f364663424154190b346354494d1876"
        ),
        random_bytes=random_bytes,
        path=AtomRequestPath.LATEST_VALUES,
    )

    sensor_data = response.parse()
    assert sensor_data is not None

    assert sensor_data["TMP"] == 29424
    assert sensor_data["HUM"] == 3375
    assert sensor_data["CO2"] == 732
    assert sensor_data["VOC"] == 277
    assert sensor_data["LUX"] == 1
    assert sensor_data["PRS"] == 6239814
    assert sensor_data["BAT"] == 2868
    assert sensor_data["TIM"] == 118
    assert sensor_data["NOI"] == 39


def test_atom_response_corentium_home_2() -> None:
    """Test the Wave Enhance request."""
    random_bytes = bytes.fromhex("CCA4")

    response = AtomResponse(
        logger=_LOGGER,
        response=bytes.fromhex(
            "1001000345CCA481A2006D32393939392F302F3331303132025831A863523234"
            + "0363523744076352333007635231591263544D501973D76348554D190D8C63424"
            + "154190B816354494D19061D"
        ),
        random_bytes=random_bytes,
        path=AtomRequestPath.LATEST_VALUES,
    )

    sensor_data = response.parse()
    assert sensor_data is not None

    assert sensor_data["TMP"] == 29655
    assert sensor_data["HUM"] == 3468
    assert sensor_data["BAT"] == 2945
    assert sensor_data["TIM"] == 1565
    assert sensor_data["R24"] == 3
    assert sensor_data["R7D"] == 7
    assert sensor_data["R30"] == 7
    assert sensor_data["R1Y"] == 18



from airthings_ble.atom.request import AtomRequest
from airthings_ble.atom.request_path import AtomRequestPath
from pytest import mark


@mark.parametrize(
    "random_bytes",
    [
        None,
        bytes.fromhex("E473"),
    ],
)
def test_atom_request(random_bytes: bytes | None) -> None:
    """Test the Wave Enhance request."""
    request = AtomRequest(url=AtomRequestPath.LATEST_VALUES, random_bytes=random_bytes)
    assert request.url == AtomRequestPath.LATEST_VALUES
    assert len(request.random_bytes) == 2
    if random_bytes is not None:
        assert request.random_bytes == random_bytes
    else:
        random_bytes = request.random_bytes

    request_bytes = request.as_bytes()

    assert request_bytes[0:2] == bytes.fromhex("0301")
    assert request_bytes[2:4] == random_bytes
    assert request_bytes[4:8] == bytes.fromhex("81A1006D")
    assert request_bytes[8:] == request.url.as_bytes()



from airthings_ble.airthings_firmware import AirthingsFirmwareVersion
from pytest import mark


def test_airthings_firmware_need_upgrade() -> None:
    fw = AirthingsFirmwareVersion(
        current_version="T-SUB-2.6.0-master+0",
        required_version="T-SUB-2.6.1-master+0",
    )
    assert fw.need_firmware_upgrade is True


def test_airthings_firmware_need_upgrade_semantic() -> None:
    fw = AirthingsFirmwareVersion(
        current_version="1.3.4",
        required_version="1.3.5",
    )
    assert fw.need_firmware_upgrade is True
    assert fw.current_version == (1, 3, 4)
    assert fw.required_version == (1, 3, 5)


def test_airthings_firmware_no_upgrade() -> None:
    fw = AirthingsFirmwareVersion(
        current_version="R-SUB-1.3.5-master+0",
        required_version="R-SUB-1.3.5-master+0",
    )
    assert fw.need_firmware_upgrade is False
    assert fw.current_version == (1, 3, 5)
    assert fw.required_version == (1, 3, 5)


def test_airthings_firmware_no_required_firmware() -> None:
    fw = AirthingsFirmwareVersion(
        current_version="T-SUB-1.0.1-master+0",
        required_version=None,
    )
    assert fw.need_firmware_upgrade is False
    assert fw.current_version == (1, 0, 1)
    assert fw.required_version is None


@mark.parametrize(
    "invalid_version",
    [
        "a.b.c",
        "1.2",
        "1.a.3",
        "invalid",
        None,
    ],
)
def test_airthings_firmware_invalid_current_version(invalid_version: str) -> None:
    fw = AirthingsFirmwareVersion(
        current_version=invalid_version,
        required_version="T-SUB-1.0.1-master+0",
    )
    assert fw.current_version is None
    assert fw.required_version == (1, 0, 1)
    assert fw.need_firmware_upgrade is False


def test_airthings_firmware_invalid_required_version() -> None:
    fw = AirthingsFirmwareVersion(
        current_version="G-SUB-1.0.1-master+0",
        required_version="invalid",
    )
    assert fw.need_firmware_upgrade is False
    assert fw.current_version == (1, 0, 1)
    assert fw.required_version is None



from airthings_ble.parser import get_radon_level


def test_radon_level() -> None:
    assert get_radon_level(0) == "very low"
    assert get_radon_level(49) == "very low"
    assert get_radon_level(50) == "low"
    assert get_radon_level(99) == "low"
    assert get_radon_level(100) == "moderate"
    assert get_radon_level(299) == "moderate"
    assert get_radon_level(300) == "high"
    assert get_radon_level(1000) == "high"



from enum import StrEnum

import cbor2


class AtomRequestPath(StrEnum):
    """Request paths for Airthings BLE Atom API"""

    LATEST_VALUES = "29999/0/31012"

    def as_cbor(self) -> bytes:
        """Get URL as bytes"""
        return cbor2.dumps(self.value)

    def as_bytes(self) -> bytes:
        """Get URL as bytes"""
        return bytes(self.value, "utf-8")



"""Airthings firmware version handling."""

import re


class AirthingsFirmwareVersion:
    """Firmware information for the Airthings device."""

    current_version: tuple[int, int, int] | None = None
    required_version: tuple[int, int, int] | None = None

    def __init__(
        self,
        current_version: str | None = None,
        required_version: str | None = None,
    ) -> None:
        self.current_version = self._version_as_tuple(current_version)
        self.required_version = self._version_as_tuple(required_version)

    def update_current_version(self, current_version: str | None) -> None:
        """Update the current version."""
        self.current_version = self._version_as_tuple(current_version)

    def update_required_version(self, required_version: str | None) -> None:
        """Update the required version."""
        self.required_version = self._version_as_tuple(required_version)

    def _version_as_tuple(self, version: str | None) -> tuple[int, int, int] | None:
        """Convert version string to tuple."""
        if not version:
            return None

        # Semantic version format: "X.Y.Z"
        if re.match(r"^\d+\.\d+\.\d+$", version):
            semantic_version = re.compile(r"(\d+)\.(\d+)\.(\d+)")
            match_obj = semantic_version.match(version)
            if not match_obj:
                return None
            major, minor, patch = match_obj.groups()
            return int(major), int(minor), int(patch)

        # Airthings version format:
        #  "T-SUB-2.6.0-master+0"
        #  "R-SUB-1.3.5-master+0"
        airthings_version = re.compile(r"^[A-Z]-SUB-(\d+)\.(\d+)\.(\d+)-.*")
        match_obj = airthings_version.match(version)
        if not match_obj:
            return None
        major, minor, patch = match_obj.groups()
        return int(major), int(minor), int(patch)

    @property
    def need_firmware_upgrade(self) -> bool:
        """Check if the device needs an update."""
        if not self.current_version or not self.required_version:
            return False

        return self.current_version < self.required_version



import os

from airthings_ble.atom.request_path import AtomRequestPath


class AtomRequest:
    """Request for Airthings BLE Atom API"""

    url: AtomRequestPath
    random_bytes: bytes

    def __init__(self, url: AtomRequestPath, random_bytes: bytes | None = None) -> None:
        self.url = url
        if random_bytes is not None:
            if len(random_bytes) != 2:
                raise ValueError("Random bytes must be exactly 2 bytes long")
            self.random_bytes = random_bytes
        else:
            self.random_bytes = os.urandom(2)

    def as_bytes(self) -> bytes:
        """Get request as bytes"""
        bytes = bytearray()
        bytes.extend(bytes.fromhex("0301"))
        bytes.extend(self.random_bytes)
        bytes.extend(bytes.fromhex("81A100"))
        bytes.extend(self.url.as_cbor())
        return bytes



import logging

from airthings_ble.parser import _decode_wave_radon

_LOGGER = logging.getLogger(__name__)


def test_wave_radon_sensor_data() -> None:
    """Test wave plus sensor data."""
    raw_data = bytearray.fromhex("013860f009001100a709ffffffffffff0000ffff")

    decoded_data = _decode_wave_radon(name="Wave2", format_type="<4B8H", scale=1.0)(
        raw_data
    )

    assert decoded_data["humidity"] == 28.0
    assert decoded_data["radon_1day_avg"] == 9
    assert decoded_data["radon_longterm_avg"] == 17
    assert decoded_data["temperature"] == 24.71



import pytest
from airthings_ble.device_type import AirthingsDeviceType


def test_device_type() -> None:
    """Test device type."""
    assert AirthingsDeviceType.WAVE_MINI.product_name == "Wave Mini"
    assert AirthingsDeviceType.WAVE_PLUS.product_name == "Wave Plus"
    assert AirthingsDeviceType.WAVE_RADON.product_name == "Wave Radon"
    assert AirthingsDeviceType.WAVE_GEN_1.product_name == "Wave Gen 1"
    assert AirthingsDeviceType.WAVE_ENHANCE_EU.product_name == "Wave Enhance"
    assert AirthingsDeviceType.WAVE_ENHANCE_US.product_name == "Wave Enhance"
    assert AirthingsDeviceType("2900") == AirthingsDeviceType.WAVE_GEN_1
    assert AirthingsDeviceType("2920") == AirthingsDeviceType.WAVE_MINI
    assert AirthingsDeviceType("2930") == AirthingsDeviceType.WAVE_PLUS
    assert AirthingsDeviceType("2950") == AirthingsDeviceType.WAVE_RADON
    assert AirthingsDeviceType("3210") == AirthingsDeviceType.WAVE_ENHANCE_EU
    assert AirthingsDeviceType("3220") == AirthingsDeviceType.WAVE_ENHANCE_US
    with pytest.raises(ValueError):
        AirthingsDeviceType("1234")

    assert AirthingsDeviceType.from_raw_value("2900") == AirthingsDeviceType.WAVE_GEN_1
    assert AirthingsDeviceType.from_raw_value("2920") == AirthingsDeviceType.WAVE_MINI
    assert AirthingsDeviceType.from_raw_value("2930") == AirthingsDeviceType.WAVE_PLUS
    assert AirthingsDeviceType.from_raw_value("2950") == AirthingsDeviceType.WAVE_RADON
    assert (
        AirthingsDeviceType.from_raw_value("3210")
        == AirthingsDeviceType.WAVE_ENHANCE_EU
    )
    assert (
        AirthingsDeviceType.from_raw_value("3220")
        == AirthingsDeviceType.WAVE_ENHANCE_US
    )

    unknown_device = AirthingsDeviceType.from_raw_value("1234")
    assert unknown_device == AirthingsDeviceType.UNKNOWN
    assert unknown_device.product_name == "Unknown"
    assert unknown_device.raw_value == "1234"


def test_battery_calculation() -> None:
    """Test battery calculation for all devices."""
    # Starting with the Wave Mini, since it has a different battery voltage range.
    # Max voltage is 4.5V, min voltage is 2.4V.

    # Starting with 5V, which is more than the max voltage
    assert AirthingsDeviceType.WAVE_MINI.battery_percentage(5.0) == 100
    assert AirthingsDeviceType.WAVE_MINI.battery_percentage(4.5) == 100
    assert AirthingsDeviceType.WAVE_MINI.battery_percentage(4.2) == 85
    assert AirthingsDeviceType.WAVE_MINI.battery_percentage(3.9) == 62
    assert AirthingsDeviceType.WAVE_MINI.battery_percentage(3.75) == 42
    assert AirthingsDeviceType.WAVE_MINI.battery_percentage(3.3) == 23
    assert AirthingsDeviceType.WAVE_MINI.battery_percentage(2.4) == 0
    assert AirthingsDeviceType.WAVE_MINI.battery_percentage(2.3) == 0

    # Repeat for the rest of the devices since they have the same amount of batteries.
    # Max voltage is 3.0V, min voltage is 2.1V.
    assert AirthingsDeviceType.WAVE_PLUS.battery_percentage(3.2) == 100
    assert AirthingsDeviceType.WAVE_PLUS.battery_percentage(3.0) == 100
    assert AirthingsDeviceType.WAVE_PLUS.battery_percentage(2.8) == 81
    assert AirthingsDeviceType.WAVE_PLUS.battery_percentage(2.6) == 53
    assert AirthingsDeviceType.WAVE_PLUS.battery_percentage(2.5) == 28
    assert AirthingsDeviceType.WAVE_PLUS.battery_percentage(2.2) == 5
    assert AirthingsDeviceType.WAVE_PLUS.battery_percentage(2.1) == 0
    assert AirthingsDeviceType.WAVE_PLUS.battery_percentage(2.0) == 0

    assert AirthingsDeviceType.WAVE_RADON.battery_percentage(3.2) == 100
    assert AirthingsDeviceType.WAVE_RADON.battery_percentage(3.0) == 100
    assert AirthingsDeviceType.WAVE_RADON.battery_percentage(2.8) == 81
    assert AirthingsDeviceType.WAVE_RADON.battery_percentage(2.6) == 53
    assert AirthingsDeviceType.WAVE_RADON.battery_percentage(2.5) == 28
    assert AirthingsDeviceType.WAVE_RADON.battery_percentage(2.2) == 5
    assert AirthingsDeviceType.WAVE_RADON.battery_percentage(2.1) == 0
    assert AirthingsDeviceType.WAVE_RADON.battery_percentage(2.0) == 0

    assert AirthingsDeviceType.WAVE_GEN_1.battery_percentage(3.2) == 100
    assert AirthingsDeviceType.WAVE_GEN_1.battery_percentage(3.0) == 100
    assert AirthingsDeviceType.WAVE_GEN_1.battery_percentage(2.8) == 81
    assert AirthingsDeviceType.WAVE_GEN_1.battery_percentage(2.6) == 53
    assert AirthingsDeviceType.WAVE_GEN_1.battery_percentage(2.5) == 28
    assert AirthingsDeviceType.WAVE_GEN_1.battery_percentage(2.2) == 5
    assert AirthingsDeviceType.WAVE_GEN_1.battery_percentage(2.1) == 0
    assert AirthingsDeviceType.WAVE_GEN_1.battery_percentage(2.0) == 0

    assert AirthingsDeviceType.WAVE_ENHANCE_EU.battery_percentage(3.2) == 100
    assert AirthingsDeviceType.WAVE_ENHANCE_EU.battery_percentage(3.0) == 100
    assert AirthingsDeviceType.WAVE_ENHANCE_EU.battery_percentage(2.8) == 81
    assert AirthingsDeviceType.WAVE_ENHANCE_EU.battery_percentage(2.6) == 53
    assert AirthingsDeviceType.WAVE_ENHANCE_EU.battery_percentage(2.5) == 28
    assert AirthingsDeviceType.WAVE_ENHANCE_EU.battery_percentage(2.2) == 5
    assert AirthingsDeviceType.WAVE_ENHANCE_EU.battery_percentage(2.1) == 0
    assert AirthingsDeviceType.WAVE_ENHANCE_EU.battery_percentage(2.0) == 0



import logging

from airthings_ble.parser import WaveRadonAndPlusCommandDecode, _decode_wave_plus

_LOGGER = logging.getLogger(__name__)


def test_wave_plus_command_decode() -> None:
    """Test wave plus command decode."""
    decode = WaveRadonAndPlusCommandDecode()
    assert decode.decode_data(
        logger=_LOGGER,
        raw_data=bytearray.fromhex(
            "6d00600c04000100008211ff00000000c04c20001f3560007006B80B0900"
        ),
    ) == {"battery": 3.0}


def test_wave_plus_sensor_data() -> None:
    """Test wave plus sensor data."""
    raw_data = bytearray.fromhex("01380d800b002200bd094cc31d036c0000007d05")

    decoded_data = _decode_wave_plus(name="Plus", format_type="<4B8H", scale=1.0)(
        raw_data
    )

    assert decoded_data["humidity"] == 28.0
    assert decoded_data["radon_1day_avg"] == 11
    assert decoded_data["radon_longterm_avg"] == 34
    assert decoded_data["temperature"] == 24.93
    assert decoded_data["voc"] == 108
    assert decoded_data["co2"] == 797
    assert decoded_data["illuminance"] == 5
    assert decoded_data["pressure"] == 999.92
